@@ -1,4 +1,5 @@
 # -*- coding: utf-8 -*-
+from .experimental import *  # noqa
 from .credentials import *  # noqa
 from .driver import *  # noqa
 from .table import *  # noqa
@@ -7,3 +8,6 @@ from .types import *  # noqa
 from .scheme import *  # noqa
 from .settings import *  # noqa
 from .resolver import *  # noqa
+from .export import *  # noqa
+from .s3list import *  # noqa
+from .auth_helpers import *  # noqa
