@@ -27,11 +27,26 @@ class DescribeTableSettings(settings.BaseRequestSettings):
     def __init__(self):
         super(DescribeTableSettings, self).__init__()
         self.include_shard_key_bounds = False
+        self.include_table_stats = False
 
     def with_include_shard_key_bounds(self, value):
         self.include_shard_key_bounds = value
         return self
 
+    def with_include_table_stats(self, value):
+        self.include_table_stats = value
+        return self
+
+
+class ExecDataQuerySettings(settings.BaseRequestSettings):
+    def __init__(self):
+        super(ExecDataQuerySettings, self).__init__()
+        self.keep_in_cache = True
+
+    def with_keep_in_cache(self, value):
+        self.keep_in_cache = value
+        return self
+
 
 class KeyBound(object):
     __slots__ = ('_equal', 'value', 'type')
@@ -96,6 +111,9 @@ class Column(object):
         self._name = name
         self._type = type
 
+    def __eq__(self, other):
+        return self.name == other.name and self._type.item == other.type.item
+
     @property
     def name(self):
         return self._name
@@ -168,6 +186,15 @@ class CompactionPolicy(object):
         return self._pb
 
 
+class SplitPoint(object):
+    def __init__(self, *args):
+        self._value = tuple(args)
+
+    @property
+    def value(self):
+        return self._value
+
+
 class ExplicitPartitions(object):
     def __init__(self, split_points):
         self.split_points = split_points
@@ -229,6 +256,27 @@ class PartitioningPolicy(object):
         return self._pb
 
 
+class TableIndex(object):
+    def __init__(self, name):
+        self._pb = _apis.ydb_table.TableIndex()
+        self._pb.name = name
+        self.name = name
+        self.index_columns = []
+
+    def with_global_index(self):
+        self._pb.global_index.SetInParent()
+        return self
+
+    def with_index_columns(self, *columns):
+        for column in columns:
+            self._pb.index_columns.append(column)
+            self.index_columns.append(column)
+        return self
+
+    def to_pb(self):
+        return self._pb
+
+
 class ReplicationPolicy(object):
     def __init__(self):
         self._pb = _apis.ydb_table.ReplicationPolicy()
@@ -261,12 +309,12 @@ class ReplicationPolicy(object):
         return self._pb
 
 
-class StorageSettings(object):
-    def __init__(self, storage_kind):
-        self.storage_kind = storage_kind
+class StoragePool(object):
+    def __init__(self, media):
+        self.media = media
 
     def to_pb(self):
-        return _apis.ydb_table.StorageSettings(storage_kind=self.storage_kind)
+        return _apis.ydb_table.StoragePool(media=self.media)
 
 
 class StoragePolicy(object):
@@ -384,11 +432,56 @@ class TableProfile(object):
         return pb
 
 
+class DateTypeColumnModeSettings(object):
+    def __init__(self, column_name, expire_after_seconds=0):
+        self.column_name = column_name
+        self.expire_after_seconds = expire_after_seconds
+
+    def to_pb(self):
+        pb = _apis.ydb_table.DateTypeColumnModeSettings()
+
+        pb.column_name = self.column_name
+        pb.expire_after_seconds = self.expire_after_seconds
+
+        return pb
+
+
+class TtlSettings(object):
+    def __init__(self):
+        self.date_type_column = None
+
+    def with_date_type_column(self, column_name, expire_after_seconds=0):
+        self.date_type_column = DateTypeColumnModeSettings(column_name, expire_after_seconds)
+        return self
+
+    def to_pb(self):
+        pb = _apis.ydb_table.TtlSettings()
+
+        if self.date_type_column is not None:
+            pb.date_type_column.MergeFrom(self.date_type_column.to_pb())
+        else:
+            raise RuntimeError("Unspecified ttl settings mode")
+
+        return pb
+
+
+class TableStats(object):
+    def __init__(self):
+        self.partitions = None
+
+    def with_partitions(self, partitions):
+        self.partitions = partitions
+        return self
+
+
 class TableDescription(object):
     def __init__(self):
         self.columns = []
         self.primary_key = []
         self.profile = None
+        self.indexes = []
+        self.ttl_settings = None
+        self.attributes = {}
 
     def with_column(self, column):
         self.columns.append(column)
@@ -408,10 +501,27 @@ class TableDescription(object):
             self.with_primary_key(pk)
         return self
 
+    def with_indexes(self, *indexes):
+        for index in indexes:
+            self.with_index(index)
+        return self
+
+    def with_index(self, index):
+        self.indexes.append(index)
+        return self
+
     def with_profile(self, profile):
         self.profile = profile
         return self
 
+    def with_ttl(self, ttl_settings):
+        self.ttl_settings = ttl_settings
+        return self
+
+    def with_attributes(self, attributes):
+        self.attributes = attributes
+        return self
+
 
 @six.add_metaclass(abc.ABCMeta)
 class AbstractTransactionModeBuilder(object):
@@ -480,27 +590,48 @@ class StaleReadOnly(AbstractTransactionModeBuilder):
         return self._name
 
 
+class BackoffSettings(object):
+    def __init__(self, ceiling=6, slot_duration=.001, uncertain_ratio=0.5):
+        self.ceiling = ceiling
+        self.slot_duration = slot_duration
+        self.uncertain_ratio = uncertain_ratio
+
+    def calc_timeout(self, retry_number):
+        slots_count = 1 << min(retry_number, self.ceiling)
+        max_duration_ms = slots_count * self.slot_duration * 1000.
+        # duration_ms = random.random() * max_duration_ms * uncertain_ratio) + max_duration_ms * (1 - uncertain_ratio)
+        duration_ms = max_duration_ms * (random.random() * self.uncertain_ratio + 1. - self.uncertain_ratio)
+        return duration_ms / 1000.
+
+
 class RetrySettings(object):
     def __init__(
             self, max_retries=10,
             max_session_acquire_timeout=3,
-            on_ydb_error_callback=None, backoff_ceiling=6, backoff_slot_duration=1,
-            get_session_client_timeout=5):
+            on_ydb_error_callback=None,
+            backoff_ceiling=6,
+            backoff_slot_duration=1,
+            get_session_client_timeout=5,
+            fast_backoff_settings=None,
+            slow_backoff_settings=None):
         self.max_retries = max_retries
         self.max_session_acquire_timeout = max_session_acquire_timeout
-        self.on_ydb_error_callback = lambda e: None if on_ydb_error_callback is None else on_ydb_error_callback
-        self.backoff_ceiling = backoff_ceiling
-        self.backoff_slot_duration = backoff_slot_duration
+        self.on_ydb_error_callback = (lambda e: None) if on_ydb_error_callback is None else on_ydb_error_callback
+        self.fast_backoff = BackoffSettings(10, .005) if fast_backoff_settings is None else fast_backoff_settings
+        self.slow_backoff = BackoffSettings(backoff_ceiling, backoff_slot_duration) \
+            if slow_backoff_settings is None else slow_backoff_settings
         self.retry_not_found = True
         self.retry_internal_error = True
         self.unknown_error_handler = lambda e: None
         self.get_session_client_timeout = get_session_client_timeout
 
+    def with_fast_backoff(self, backoff_settings):
+        self.fast_backoff = backoff_settings
+        return self
 
-def calc_backoff_timeout(settings, retry_number):
-    slots_count = 1 << min(retry_number, settings.backoff_ceiling)
-    max_duration = slots_count * settings.backoff_slot_duration
-    return random.random() * max_duration
+    def with_slow_backoff(self, backoff_settings):
+        self.slow_backoff = backoff_settings
+        return self
 
 
 def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
@@ -512,7 +643,7 @@ def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
         try:
             return callee(*args, **kwargs)
         except (
-                issues.Unavailable, issues.Aborted, issues.BadSession,
+                issues.Aborted, issues.BadSession,
                 issues.NotFound, issues.InternalError) as e:
             status = e
             retry_settings.on_ydb_error_callback(e)
@@ -527,8 +658,16 @@ def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
             status = e
             retry_settings.on_ydb_error_callback(e)
             time.sleep(
-                calc_backoff_timeout(
-                    retry_settings,
+                retry_settings.slow_backoff.calc_timeout(
+                    attempt
+                )
+            )
+
+        except issues.Unavailable as e:
+            status = e
+            retry_settings.on_ydb_error_callback(e)
+            time.sleep(
+                retry_settings.fast_backoff.calc_timeout(
                     attempt
                 )
             )
@@ -545,25 +684,35 @@ def retry_operation_sync(callee, retry_settings=None, *args, **kwargs):
     raise status
 
 
+class TableClientSettings(object):
+    def __init__(self):
+        self._client_query_cache_enabled = True
+
+    def with_client_query_cache(self, enabled):
+        self._client_query_cache_enabled = enabled
+        return self
+
+
 class TableClient(object):
-    def __init__(self, driver):
+    def __init__(self, driver, table_client_settings=None):
         self._driver = driver
+        self._table_client_settings = TableClientSettings() if table_client_settings is None else table_client_settings
 
     def session(self):
-        return Session(self._driver)
+        return Session(self._driver, self._table_client_settings._client_query_cache_enabled)
 
 
 class TableSchemeEntry(scheme.SchemeEntry):
     def __init__(
             self, name, owner, type, effective_permissions, permissions, columns, primary_key, shard_key_bounds,
-            *args, **kwargs):
+            indexes, table_stats, ttl_settings, attributes, *args, **kwargs):
 
         super(TableSchemeEntry, self).__init__(name, owner, type, effective_permissions, permissions, *args, **kwargs)
         self.primary_key = [pk for pk in primary_key]
-        self.columns = [
-            Column(column.name, convert.type_to_native(column.type))
-            for column in columns
-        ]
+        self.columns = [Column(column.name, convert.type_to_native(column.type)) for column in columns]
+        self.indexes = [
+            TableIndex(index.name).with_index_columns(*tuple(col for col in index.index_columns))
+            for index in indexes]
         self.shard_key_ranges = []
         left_key_bound = None
         for shard_key_bound in shard_key_bounds:
@@ -592,57 +741,31 @@ class TableSchemeEntry(scheme.SchemeEntry):
             self.shard_key_ranges.append(
                 KeyRange(None, None))
 
+        self.ttl_settings = None
+        if ttl_settings is not None:
+            if ttl_settings.HasField('date_type_column'):
+                self.ttl_settings = TtlSettings().with_date_type_column(
+                    ttl_settings.date_type_column.column_name,
+                    ttl_settings.date_type_column.expire_after_seconds
+                )
 
-class SyncResponseIterator(object):
-    def __init__(self, it, wrapper):
-        self.it = it
-        self.wrapper = wrapper
+        self.table_stats = None
+        if table_stats is not None:
+            self.table_stats = TableStats()
+            if table_stats.partitions != 0:
+                self.table_stats = self.table_stats.with_partitions(
+                    table_stats.partitions
+                )
 
-    def cancel(self):
-        self.it.cancel()
-        return self
-
-    def __iter__(self):
-        return self
-
-    def _next(self):
-        return self.wrapper(next(self.it))
-
-    def next(self):
-        return self._next()
-
-    def __next__(self):
-        return self._next()
-
-
-class AsyncResponseIterator(object):
-    def __init__(self, it, wrapper):
-        self.it = it
-        self.wrapper = wrapper
-
-    def cancel(self):
-        self.it.cancel()
-        return self
-
-    def __iter__(self):
-        return self
-
-    def _next(self):
-        return interceptor.operate_async_stream_call(self.it, self.wrapper)
-
-    def next(self):
-        return self._next()
-
-    def __next__(self):
-        return self._next()
+        self.attributes = attributes
 
 
 class Session(object):
     __slots__ = ('_state', '_driver', '__weakref__')
 
-    def __init__(self, driver):
+    def __init__(self, driver, client_query_cache_enabled):
         self._driver = driver
-        self._state = _session_impl.SessionState()
+        self._state = _session_impl.SessionState(client_query_cache_enabled)
 
     def __lt__(self, other):
         return self.session_id < other.session_id
@@ -663,13 +786,16 @@ class Session(object):
         """
         return self._state.session_id is not None
 
+    def pending_query(self):
+        return self._state.pending_query()
+
     def reset(self):
         """
         Perform session state reset (that includes cleanup of the session_id, query cache, and etc.)
         """
         return self._state.reset()
 
-    def read_table(self, path, key_range=None, columns=(), ordered=False, row_limit=None):
+    def read_table(self, path, key_range=None, columns=(), ordered=False, row_limit=None, settings=None, use_snapshot=None):
         """
         Perform an read table request.
 
@@ -705,11 +831,11 @@ class Session(object):
         :param row_limit: (optional) A number of rows to read.
         :return: SyncResponseIterator instance
         """
-        request = _session_impl.read_table_request_factory(self._state, path, key_range, columns, ordered, row_limit)
-        stream_it = self._driver(request, _apis.TableService.Stub, _apis.TableService.StreamReadTable)
-        return SyncResponseIterator(stream_it, _session_impl.wrap_read_table_response)
+        request = _session_impl.read_table_request_factory(self._state, path, key_range, columns, ordered, row_limit, use_snapshot=use_snapshot)
+        stream_it = self._driver(request, _apis.TableService.Stub, _apis.TableService.StreamReadTable, settings=settings)
+        return _utilities.SyncResponseIterator(stream_it, _session_impl.wrap_read_table_response)
 
-    def async_read_table(self, path, key_range=None, columns=(), ordered=False, row_limit=None):
+    def async_read_table(self, path, key_range=None, columns=(), ordered=False, row_limit=None, settings=None, use_snapshot=None):
         """
         Perform an read table request.
 
@@ -748,9 +874,9 @@ class Session(object):
         """
         if interceptor is None:
             raise RuntimeError("Async read table is not available due to import issues")
-        request = _session_impl.read_table_request_factory(self._state, path, key_range, columns, ordered, row_limit)
-        stream_it = self._driver(request, _apis.TableService.Stub, _apis.TableService.StreamReadTable)
-        return AsyncResponseIterator(stream_it, _session_impl.wrap_read_table_response)
+        request = _session_impl.read_table_request_factory(self._state, path, key_range, columns, ordered, row_limit, use_snapshot=use_snapshot)
+        stream_it = self._driver(request, _apis.TableService.Stub, _apis.TableService.StreamReadTable, settings=settings)
+        return _utilities.AsyncResponseIterator(stream_it, _session_impl.wrap_read_table_response)
 
     def keep_alive(self, settings=None):
         return self._driver(
@@ -895,6 +1021,24 @@ class Session(object):
             self._state.endpoint,
         )
 
+    def explain(self, yql_text, settings=None):
+        """
+        Expiremental API.
+
+        :param yql_text:
+        :param settings:
+        :return:
+        """
+        return self._driver(
+            _session_impl.explain_data_query_request_factory(self._state, yql_text),
+            _apis.TableService.Stub,
+            _apis.TableService.ExplainDataQuery,
+            _session_impl.wrap_explain_response,
+            settings,
+            (self._state, ),
+            self._state.endpoint,
+        )
+
     def create_table(self, path, table_description, settings=None):
         """
         Create a YDB table.
@@ -970,45 +1114,75 @@ class Session(object):
         )
 
     @_utilities.wrap_async_call_exceptions
-    def async_alter_table(self, path, add_columns, drop_columns, settings=None):
+    def async_alter_table(
+            self, path,
+            add_columns=None, drop_columns=None,
+            settings=None,
+            alter_attributes=None,
+            add_indexes=None, drop_indexes=None,
+            set_ttl_settings=None, drop_ttl_settings=None):
         return self._driver.future(
-            _session_impl.alter_table_request_factory(self._state, path, add_columns, drop_columns),
+            _session_impl.alter_table_request_factory(
+                self._state, path,
+                add_columns, drop_columns,
+                alter_attributes,
+                add_indexes, drop_indexes,
+                set_ttl_settings, drop_ttl_settings
+            ),
             _apis.TableService.Stub,
             _apis.TableService.AlterTable,
-            _session_impl.wrap_operation,
+            _session_impl.AlterTableOperation,
             settings,
-            (self._state,),
+            (self._driver, ),
             self._state.endpoint,
         )
 
-    def alter_table(self, path, add_columns, drop_columns, settings=None):
+    def alter_table(
+            self, path,
+            add_columns=None, drop_columns=None,
+            settings=None,
+            alter_attributes=None,
+            add_indexes=None, drop_indexes=None,
+            set_ttl_settings=None, drop_ttl_settings=None):
         return self._driver(
-            _session_impl.alter_table_request_factory(self._state, path, add_columns, drop_columns),
+            _session_impl.alter_table_request_factory(
+                self._state, path,
+                add_columns, drop_columns,
+                alter_attributes,
+                add_indexes, drop_indexes,
+                set_ttl_settings, drop_ttl_settings
+            ),
             _apis.TableService.Stub,
             _apis.TableService.AlterTable,
-            _session_impl.wrap_operation,
+            _session_impl.AlterTableOperation,
             settings,
-            (self._state,),
+            (self._driver, ),
             self._state.endpoint,
         )
 
-    @_utilities.wrap_async_call_exceptions
     def async_copy_table(self, source_path, destination_path, settings=None):
-        return self._driver.future(
-            _session_impl.copy_table_request_factory(self._state, source_path, destination_path),
-            _apis.TableService.Stub,
-            _apis.TableService.CopyTable,
-            _session_impl.wrap_operation,
-            settings,
-            (self._state,),
-            self._state.endpoint,
-        )
+        return self.async_copy_tables([(source_path, destination_path)], settings=settings)
 
     def copy_table(self, source_path, destination_path, settings=None):
-        return self._driver(
-            _session_impl.copy_table_request_factory(self._state, source_path, destination_path),
+        return self.copy_tables([(source_path, destination_path)], settings=settings)
+
+    @_utilities.wrap_async_call_exceptions
+    def async_copy_tables(self, source_destination_pairs, settings=None):
+        return self._driver.future(
+            _session_impl.copy_tables_request_factory(self._state, source_destination_pairs),
             _apis.TableService.Stub,
-            _apis.TableService.CopyTable,
+            _apis.TableService.CopyTables,
+            _session_impl.wrap_operation,
+            settings,
+            (self._state,),
+            self._state.endpoint,
+        )
+
+    def copy_tables(self, source_destination_pairs, settings=None):
+        return self._driver(
+            _session_impl.copy_tables_request_factory(self._state, source_destination_pairs),
+            _apis.TableService.Stub,
+            _apis.TableService.CopyTables,
             _session_impl.wrap_operation,
             settings,
             (self._state,),
@@ -1123,11 +1297,12 @@ class TxContext(object):
         :param query: A query: YQL text or DataQuery instance. E
         :param parameters: A dictionary with parameters values.
         :param commit_tx: A special flag that allows transaction commit
-        :param settings: An additional request settings
+        :param settings: A request settings (an instance of ExecDataQuerySettings)
         :return: A future of query execution
         """
         return self._driver.future(
-            _tx_ctx_impl.execute_request_factory(self._session_state, self._tx_state, query, parameters, commit_tx),
+            _tx_ctx_impl.execute_request_factory(
+                self._session_state, self._tx_state, query, parameters, commit_tx, settings),
             _apis.TableService.Stub,
             _apis.TableService.ExecuteDataQuery,
             _tx_ctx_impl.wrap_result_and_tx_id,
@@ -1148,7 +1323,8 @@ class TxContext(object):
         :return: A result sets or exception in case of execution errors
         """
         return self._driver(
-            _tx_ctx_impl.execute_request_factory(self._session_state, self._tx_state, query, parameters, commit_tx),
+            _tx_ctx_impl.execute_request_factory(
+                self._session_state, self._tx_state, query, parameters, commit_tx, settings),
             _apis.TableService.Stub,
             _apis.TableService.ExecuteDataQuery,
             _tx_ctx_impl.wrap_result_and_tx_id,
@@ -1289,7 +1465,7 @@ class TxContext(object):
 
 
 class SessionPool(object):
-    def __init__(self, driver, size=100, workers_threads_count=4, initializer=None):
+    def __init__(self, driver, size=100, workers_threads_count=4, initializer=None, min_pool_size=0):
         """
         An object that encapsulates session creation, deletion and etc. and maintains
         a pool of active sessions of specified size
@@ -1305,18 +1481,38 @@ class SessionPool(object):
                     "To prepare statements in session use keep in cache feature."
                 )
 
-        self._pool_impl = _sp_impl.SessionPoolImpl(driver, size, workers_threads_count, initializer)
+        self._pool_impl = _sp_impl.SessionPoolImpl(driver, size, workers_threads_count, initializer, min_pool_size)
 
     def retry_operation_sync(self, callee, retry_settings=None, *args, **kwargs):
 
         retry_settings = RetrySettings() if retry_settings is None else retry_settings
 
         def wrapped_callee():
-            with self.checkout(blocking=False, timeout=retry_settings.get_session_client_timeout) as session:
+            with self.checkout(timeout=retry_settings.get_session_client_timeout) as session:
                 return callee(session, *args, **kwargs)
 
         return retry_operation_sync(wrapped_callee, retry_settings)
 
+    @property
+    def active_size(self):
+        return self._pool_impl.active_size
+
+    @property
+    def free_size(self):
+        return self._pool_impl.free_size
+
+    @property
+    def busy_size(self):
+        return self._pool_impl.busy_size
+
+    @property
+    def max_size(self):
+        return self._pool_impl.max_size
+
+    @property
+    def waiters_count(self):
+        return self._pool_impl.waiters_count
+
     def subscribe(self):
         return self._pool_impl.subscribe()
 
@@ -1329,6 +1525,17 @@ class SessionPool(object):
     def release(self, session):
         return self._pool_impl.put(session)
 
+    def async_checkout(self):
+        """
+        Returns a context manager that asynchronously checkouts a session from the pool.
+
+        with my_pool.async_checkout() as async_session:
+            session = await async_session
+            results = await session.transaction().async_execute(....)
+
+        """
+        return AsyncSessionCheckout(self)
+
     def checkout(self, blocking=True, timeout=None):
         return SessionCheckout(self, blocking, timeout)
 
@@ -1342,6 +1549,26 @@ class SessionPool(object):
         self.stop()
 
 
+class AsyncSessionCheckout(object):
+    __slots__ = ('subscription', 'pool')
+
+    def __init__(self, pool):
+        """
+        A context manager that asynchronously checkouts a session for the specified pool
+        and returns it on manager exit.
+        :param pool: A SessionPool instance.
+        """
+        self.pool = pool
+        self.subscription = None
+
+    def __enter__(self):
+        self.subscription = self.pool.subscribe()
+        return self.subscription
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.pool.unsubscribe(self.subscription)
+
+
 class SessionCheckout(object):
     __slots__ = ('_acquired', '_pool', '_blocking', '_timeout')
 
